(* List stuff imported from list.apl *)

(* append(A,B,C) : append A to head of B *)
pred append([A],[A],[A]).
append([],L,L).
append([X|L1],L2,[X|L3]) :- append(L1,L2,L3).

pred remove(A,[A],[A]).
remove(X,[],[]).
remove(X,[X|L],L')     :- remove(X,L,L').
remove(X,[Y|L],[Y|L']) :- remove(X,L,L'),not(X=Y).

pred mem(A,[A]).
mem(X,[X|L]).
mem(X,[Y|L]) :- mem(X,L).

pred reverse([A],[A]).
pred rev([A],[A],[A]).
rev(L,[],L).
rev(L,X::L',L''):- rev(X::L,L',L'').
reverse(L,M)    :- rev([],L,M).

pred concat([[A]],[A]).
concat([],[]).
concat([A|L],L') :- concat(L,L''), append(A,L'',L').




id : name_type.
tid : name_type.
type label = string.
tm : type.
tp : type.
pattern : type.
pterm : type.


(* terms *)
var : id -> tm.                          (* variable *)
lam : id\(tp,tm) -> tm.                  (* abstraction *)
app : tm -> tm -> tm.                    (* application *)
tlam : tid\(tp,tm) -> tm.                (* type abstraction *)
tapp : tm -> tp -> tm.                   (* term abstraction *)
rcd : [(label,tm)] -> tm.                (* record *)
proj : tm -> label -> tm.                (* projection *)
patt : tm -> pattern -> pterm -> tm.     (* pattern binding *)

(* for representing patterns, we have to be clever to get the variables bound properly. 

   A pattern such as: 

      let {l1 = {m1 = y1 : S1, m2 = y2 : S2}, l2 = x2 : T2} = E1 in y1 (x2 y2)

   is encoded as: 

      patt E1 (prcd [(l1, prcd [(m1,pvar S1),(m2,pvar S2)]), (l2,pvar T2)]) 
              (pterm2(y1\(S1,pterm2(y2\(S2,pterm1(x2\(T2,app (var y1) (app (var x2) (var y2))))))))

   So the variables in the pattern must be flattened so they can be bound properly in E2, 
   but then we have to do all sorts of extra stuff to make sure the patt encoding is 
   well-formed etc. 

*)

(* patterns *)
pvar : tp -> pattern.                    (* variable pattern *)
prcd : [(label,pattern)] -> pattern.     (* record pattern *)

(* pterms used for flattening variables to be bound from the pattern *)
pterm1 : id\(tp,tm) -> pterm. 
pterm2 : id\(tp,pterm) -> pterm. 


(* types *)
tvar: tid -> tp.              (* type variable *)
top : tp.                     (* maximum type  *)
==> : tp -> tp -> tp.         (* type of functions *)
tall : tid\(tp,tp) -> tp.     (* universal type *)
trcd : [(label,tp)] -> tp.    (* record type *)

infixr ==> 5.


(* contexts *)
pair : type.
id_pair : id -> tp -> pair.    (* term variable binding *)
tid_pair : tid -> tp -> pair.  (* type variable binding *)
type ctxt = [pair].


(* Do all this subst stuff for rcdords and patterns *)
(* substitution of a term into a term*)
pred subst_patt (id\pterm) tm pterm.
pred subst_list (id\[(label,tm)]) tm [(label,tm)].
pred subst (id\tm) tm tm.
subst (a\var a) E E.
subst (a\var B) E (var B) :- a # B.
subst (a\lam (b\(T1,E1))) E (lam (b\(T1,E1'))) :- b # E, subst (a\E1) E E1'.
subst (a\app E1 E2) E (app E1' E2') :- subst (a\E1) E E1', subst (a\E2) E E2'.
subst (a\tlam (b\(T,E1))) E (tlam (b\(T,E1'))) :- subst (a\E1) E E1'.
subst (a\tapp E1 T1) E (tapp E1' T1) :- subst (a\E1) E E1'.
subst (a\rcd L) E (rcd L') :- subst_list (a\L) E L'.
subst (a\proj T L) E (proj T' L) :- subst (a\T) E T'.
subst (a\patt T1 P Pt) E (patt T1' P Pt')  :- 
        subst (a\T1) E T1', subst_patt (a\Pt) E Pt'. 

subst_patt (a\(pterm1 (b\(T,E1)))) E (pterm1 (b\(T,E1'))) :-
        b # E, subst (a\E1) E E1'.
subst_patt (a\(pterm2 (b\(T,Pt)))) E (pterm2 (b\(T,Pt'))) :-
        b # E, subst_patt (a\Pt) E Pt'.

subst_list (a\[]) E [].
subst_list (a\[(L,E1)|Es]) E ([(L,E1')|Es']) :- subst (a\E1) E E1', subst_list (a\Es) E Es'.

(* substitute type into a type *)
pred subst_tp_tp_list (tid\[(label,tp)]) tp [(label,tp)].
pred subst_tp_tp (tid\tp) tp tp.
subst_tp_tp (a\tvar a) T T.
subst_tp_tp (a\tvar B) T (tvar B) :- a # B.
subst_tp_tp (a\top) T top.
subst_tp_tp (a\(T1 ==> T2)) T (T1' ==> T2') :- 
	subst_tp_tp (a\T1) T T1', subst_tp_tp (a\T2) T T2'.
subst_tp_tp (a\tall(b\(T1,T2))) T (tall (b\(T1',T2'))) :-
	b # T, subst_tp_tp (a\T1) T T1', subst_tp_tp (a\T2) T T2'.
subst_tp_tp (a\trcd L) T (trcd L') :- subst_tp_tp_list (a\L) T L'.

subst_tp_tp_list (a\[]) T [].
subst_tp_tp_list (a\[(L,T1)|Ts]) T ([(L,T1')|Ts']) :- 
        subst_tp_tp (a\T1) T T1', subst_tp_tp_list (a\Ts) T Ts'.

(* substitute a type into a term *)
pred subst_tp_tm (tid\tm) tp tm.
pred subst_tp_tm_patt (tid\pattern) tp pattern. 
pred subst_tp_tm_patt_list (tid\[(label,pattern)]) tp [(label,pattern)].
pred subst_tp_tm_pterm (tid\pterm) tp pterm.
pred subst_tp_tm_list (tid\[(label,tm)]) tp [(label,tm)].
subst_tp_tm (a\var B) T (var B).
subst_tp_tm (a\lam (b\(T1,E1))) T (lam (b\(T1',E1'))) :-
	subst_tp_tp (a\T1) T T1', subst_tp_tm (a\E1) T E1'.
subst_tp_tm (a\app E1 E2) T (app E1' E2') :- 
	subst_tp_tm (a\E1) T E1', subst_tp_tm (a\E2) T E2'.
subst_tp_tm (a\tlam(b\(T1,E1))) T (tlam(b\(T1',E1'))) :-
	b # T, subst_tp_tm (a\E1) T E1', subst_tp_tp (a\T1) T T1'.
subst_tp_tm (a\tapp E1 T1) T (tapp E1' T1') :-
	subst_tp_tm (a\E1) T E1', subst_tp_tp (a\T1) T T1'.
subst_tp_tm (a\rcd L) T (rcd L') :- subst_tp_tm_list (a\L) T L'.
subst_tp_tm (a\proj E L) T (proj E' L) :- subst_tp_tm (a\E) T E'.
subst_tp_tm (a\patt E1 P Pt) T (patt E1' P' Pt') :-
        subst_tp_tm (a\E1) T E1', subst_tp_tm_patt (a\P) T P', 
	subst_tp_tm_pterm (a\Pt) E Pt'.

subst_tp_tm_patt (a\pvar T1) T (pvar T1') :- subst_tp_tp (a\T1) T T1'.
subst_tp_tm_patt (a\prcd Ls) T (prcd Ls') :- subst_tp_tm_patt_list (a\Ls) T Ls'.

subst_tp_tm_patt_list (a\[]) T [].
subst_tp_tm_patt_list (a\[(L1,P1)|Ls]) T ([(L1,P1')|Ls']) :- 
        subst_tp_tm_patt (a\P1) T P1', subst_tp_tm_patt_list (a\Ls) T Ls'.

subst_tp_tm_pterm (a\pterm1(b\(T1,E1))) T (pterm1(b\(T1',E1'))) :-
        subst_tp_tp (a\T1) T T1', subst_tp_tm (a\E1) T E1'.
subst_tp_tm_pterm (a\pterm2(b\(T1,Pt1))) T (pterm2(b\(T1',Pt1'))) :-
        subst_tp_tp (a\T1) T T1', subst_tp_tm_pterm (a\Pt1) T Pt1'.

subst_tp_tm_list (a\[]) T [].
subst_tp_tm_list (a\[(L,E)|Es]) T ([(L,E')|Es']) :-
        subst_tp_tm (a\E) T E', subst_tp_tm_list (a\Es) T Es'.

(* values *)

pred value tm. 
value (var X).
value (lam (a\(T,E))).
value (tlam (a\(T,E))).
value (rcd []). 
value (rcd [(Li,T)|L]) :- value T, value (rcd L).

pred value_list [(label,tm)]. 
value_list []. 
value_list [(Li,A)|As] :- value A, value_list As. 


(* subtyping *)
pred suba ctxt tp tp.

(* some auxiliary predicates *)
pred suba_rcd ctxt [(label,tp)] [(label,tp)].
pred mem_label_tp ctxt label tp [(label,tp)].


(* SA-TOP *)
suba G S top.
(* SA-REFL-TVAR *)
suba G X X.
(* SA-TRANS-TVAR *)
suba G (tvar X) T :- mem((tid_pair X U),G), suba G U T.
(* SA-ARROW *)
suba G (S1 ==> S2) (T1 ==> T2) :- suba G T1 S1, suba G S2 T2.
(* SA-ALL *)
suba G (tall (x\(S1,S2))) (tall (x\(T1,T2))) :-
	x # G, suba G S1 T1, suba ((tid_pair x T1)::G) S2 T2.
(* SA-RCD *)
suba G (trcd K) (trcd L) :- suba_rcd G K L.

(* auxiliaries for record subtyping *)

(* find the pair labelled by Li in the record and see if the subtyping holds there *)
mem_label_tp G Li Ti [(Li,Si)|K'] :- suba G Si Ti.
mem_label_tp G Li Ti [(Ki,Si)|K'] :- mem_label_tp G Li Ti K'.

(* record subtyping : SA-RCD *)
suba_rcd G K [].
suba_rcd G K [(Li,Ti)|L'] :- 
        mem_label_tp G Li Ti K, suba_rcd G K L'.




(* typing *)
pred typeof ctxt tm tp.

(* auxiliary predicates used for subtyping *)
pred typeof_rcd ctxt [(label,tm)] [(label,tp)].
pred typeof_proj [(label,tp)] label tp.
pred typeof_patt ctxt pattern pterm tp ctxt tm.
pred typeof_patt_match ctxt pattern tp.
pred typeof_patt_match_list ctxt [(label,pattern)] [(label,tp)].
pred typeof_flat_patt ctxt pterm tp ctxt tm.
pred extract_patt_tm pterm tm. 
pred wf_patt ctxt pattern pterm.
pred first_type tp tp tp.
pred find_pml label [(label,tp)] tp.
pred is_trcd tp.

(* T-Var *)
typeof G (var X) T :- mem((id_pair X T),G).              

(* T-APP *)
typeof G (app E1 E2) T12 :-
	typeof G E1 (T11 ==> T12), 
	typeof G E2 T11.

(* T-TAPP *)
typeof G (tapp E1 T2) T12' :-
	typeof G E1 (tall (x\(T11,T12))), 
	suba G T2 T11,
	subst_tp_tp (x\T12) T2 T12'.

(* T-ABS *)
typeof G (lam(x\(T1,E))) (T1 ==> T2) :-
	x # G, typeof ((id_pair x T1)::G) E T2.

(* T-TABS *)
typeof G (tlam(x\(T1,E))) (tall(x\(T1,T2))) :-
	x # G, typeof ((tid_pair x T1)::G) E T2.


(* T-RCD *)
typeof G (rcd Ls) (trcd Ts) :- 
        typeof_rcd G Ls Ts. 

(* T-PROJ *)
typeof G (proj E L) T :- 
        typeof G E (trcd Ts), typeof_proj Ts L T. 

(* T-LET *)
(* typeof G E1 T1 --- G |- E1 T1 
   typeof_patt P Pt T1 Delta --- |- (P and Pt) : T1 => Delta 
   typeof G' E2 T2 --- G,Delta |- E2 : T2   (using append(G,Delta,G')) *)

typeof G (patt E1 P Pt) T2 :- 
        typeof G E1 T1, typeof_patt G P Pt T1 Delta E2, append(G,Delta,G'),
	typeof G' E2 T2.

(* T-SUB *)
typeof G E T :- 
	typeof G E S, suba G S T, not (S=T).




(* Record typing *)
typeof_rcd G [] [].
typeof_rcd G [(Li,Ei)|L'] [(Li,Ti)|T'] :- typeof G Ei Ti, typeof_rcd G L' T'. 

(* Find the matching label in the record *)
typeof_proj [(Li,Ti)|L'] Li Ti.
typeof_proj [(Lj,Tj)|L'] Li Ti :- typeof_proj L' Li Ti.

(* pattern P has type T and pterm Pt with type T gives typing environment D (second premise of T-LET ) *)
typeof_patt G P Pt T D E2 :- typeof_patt_match G P T, typeof_flat_patt G Pt T D E2.

(* Type of the pattern *)
typeof_patt_match G (pvar T) T.
typeof_patt_match G (pvar (tvar X)) T' :- mem((tid_pair X T'), G).
typeof_patt_match G (prcd L) (trcd T) :- typeof_patt_match_list G L T.

typeof_patt_match_list G [] [].
typeof_patt_match_list G [] Ls.
typeof_patt_match_list G ([(Ln,Lt)|Ls]) ([(Ln,Tt)|Ts]) :- typeof_patt_match G Lt Tt, typeof_patt_match_list G Ls Ts.
typeof_patt_match_list G ([(Kn,Lt)|Ls]) Ts :- find_pml Kn Ts Tt, typeof_patt_match G Lt Tt, typeof_patt_match_list G Ls Ts. 

find_pml Kn [(Kn,Tt)|Ks] Tt.
find_pml Kn [(Ln,Tn)|Ls] Tt :- find_pml Kn Ls Tt.

(* Pattern typing rules used to get a typing environment from a pterm *)
typeof_flat_patt G (pterm1	 (x\(T,E))) T' ([(id_pair x T)]) E :- x # G. 
typeof_flat_patt G (pterm1 (x\((tvar t),E))) T ([(id_pair x T)]) E :- x # G, mem((tid_pair t T), G).
typeof_flat_patt G (pterm2 (x\(T,Pt))) Tr Ds E :- 
         typeof_flat_patt G Pt Tr' Ds' E, append(Ds',[(id_pair x T)],Ds), x # G, x # Ds', first_type Tr T Tr'.     

(* Strip away the binders and just return the expression E *)
extract_patt_tm (pterm1(x\(T,E))) E. 
extract_patt_tm (pterm2(x\(T,P))) E :- extract_patt_tm P E. 

(* Get the first type in the record type, depth-first *)
first_type (trcd [(L,T)|Ls]) T (trcd Ls). 
first_type (trcd [(L,T1)|Ls]) T (trcd [(L,T1')|Ls]) :- is_trcd T1, first_type T1 T T1'.

is_trcd (trcd T).

(* ?typeof [] (tlam (xt \ (top,(tlam (yt \ top,(lam (x \ (xt, lam (y \ (yt, (tlam (rt \ top, ( lam (p \ (arr xt (arr yt rt), app (app p x) y))))))))))))))). *)

(* ?typeof [] (tlam (xt \ (top,(tlam (yt \ (top,(lam (x \ ((tvar xt),(lam (y \ (tvar yt,(tlam (rt \ (top,(lam (p \ ((tvar xt) ==> ((tvar yt) ==> (tvar rt)), app (app (var p) (var x)) (var y))))))))))))))))))) T. *)

(* evaluation *)
pred ---> tm tm. 
infixr ---> 5.

(* auxiliaries for evaluation *)
pred match pterm [tm] tm.
pred flatten_rcdord_val pattern tm [tm].
pred is_rcdord tm. 
pred flatten_rcdord_val_list [(label,pattern)] [(label,tm)] [tm].
pred eval_list [(label,tm)] [(label,tm)].
pred find_li label [(label,tm)] tm.

(* E-APPABS *)
(app (lam(x\(T11,E12))) V2) ---> E' :- value V2, subst (x\E12) V2 E'.
(* E-TAPPTABS *)
(tapp (tlam (x\(T11,E12))) T2) ---> E':- subst_tp_tm (x\T12) T2 E'.
(* E-PROJRCD *)
(proj (rcd L) Li) ---> E' :- mem((Li,E'), L), value_list L. 
(* E-LETV *)
(patt V1 P Pt) ---> E' :- value V1, flatten_rcdord_val P V1 V1',  match Pt V1' E'.



(* E-APP1 *)
(app E E2) ---> (app E' E2) :- E ---> E'. 
(* E-APP2 *)
(app V E) ---> (app V E') :- E ---> E', value V. 

(* E-TAPP1 *)
(tapp E T) ---> (tapp E' T) :- E ---> E'. 
(* E-PROJ1 *)
(proj E Li) ---> (proj E' Li) :- E ---> E'. 
(* E-RCD1 *)
(rcd L) ---> (rcd L') :- eval_list L L'.
(* E-LET1 *)
(patt E P Pt) ---> (patt E' P Pt) :- E ---> E'. 


(* flatten the record value in to a list of values *)
flatten_rcdord_val (prcd P) (rcd L) V1' :- flatten_rcdord_val_list P L V1'. 
flatten_rcdord_val (pvar P) V1 [V1] :- not (is_rcdord V1).

flatten_rcdord_val_list ([]) V []. 
flatten_rcdord_val_list ([(Li,Pi)|Ps]) ([(Li,Vi)|Ls]) Ls' :- 
	flatten_rcdord_val_list Ps Ls Ls'', flatten_rcdord_val Pi Vi Vis, append(Vis,Ls'',Ls').

flatten_rcdord_val_list ([(Li,Pi)|Ps]) ([(Ki,Vi)|Ls]) Ls' :- 
       flatten_rcdord_val_list Ps ([(Ki,Vi)|Ls]) Ls'', (* is_rcdord Vi, *)
       find_li Li Ls V, flatten_rcdord_val Pi V Vis, append(Vis,Ls'',Ls').

(* find a particular label in a list *)
find_li Li ([(Li,Vi)|Ls]) Vi. 
find_li Li ([(Ki,Vi)|Ls]) V :- find_li Li Ls V. 

is_rcdord (rcd L).


(* similar to the matching rules but for flattened records *)
match (pterm1 (x\(T,E2))) ([V|Vs]) E2' :- subst (x\E2) V E2'.                        
match (pterm2 (x\(T,Pt))) ([V|Vs]) E2' :- subst_patt (x\Pt) V Pt', match Pt' Vs E2'. 

(* evaluate the first non-value field of a record *)
eval_list ([]) ([]). 
eval_list ([(Li,A)|As]) ([(Li,A)|Bs]) :- value A, eval_list As Bs. 
eval_list ([(Li,A)|As]) ([(Li,B)|As]) :- A ---> B. 



(* reflexive-transitive closure of ---> *)
pred ===> tm tm. 
infixr ===> 5. 
E ===> E :- value E.
E ===> E'' :- E ---> E', E' ===> E''.


?(app (lam (x\ (top, var x))) (lam (y\ (top, var y)))) ---> E. 

?(proj (rcd [("l1", (var x)), ("l2", rcd [("m1", (var y)), ("m2", (var z))])]) "l2") ===> rcd [("m1", (var y)), ("m2", (var z))].

?(proj (proj (rcd [("l1", (var x)), ("l2", rcd [("m1", (var y)), ("m2", (var z))])]) "l2") "m1") ===>  (var y).

? typeof [(id_pair x top)] (rcd [("l1", (var x))]) T.

? typeof [(id_pair x top), (id_pair y (top ==> top)), (id_pair z (top))] (rcd [("l1", (var x)), ("l2", rcd [("m1", (var y)), ("m2", (var z))])]) T. 

? typeof [(id_pair x top), (id_pair y top), (id_pair z top)] (proj (rcd [("l1", (var x)), ("l2", rcd [("m1", (var y)), ("m2", (var z))])]) "l2")  T.

? (patt (rcd [("l1", rcd [("m1", (var a)), ("m2", (var b))]), ("l2", (var c))]) (prcd [("l1", prcd [("m1",pvar top),("m2",pvar top)]), ("l2",pvar top)]) (pterm2(y1\(top,pterm2(y2\(top,pterm1(x2\(top,var y2)))))))) ===> E. 

? (patt (rcd [("l1", var x), ("l2", var y)]) (prcd [("l1", pvar top), ("l2", pvar top)]) (pterm2 (y1\(top, pterm1 (y2\(top, var y2)))))) ===> E. 

? typeof [(id_pair x (tall (y\(top, tvar y))))] (var x) T. 

? typeof [(id_pair x (tall (y\(top, tvar y))))] (tapp (var x) top) T.

? typeof [] (lam (x\(top,var x))) T. 

? typeof [] (tlam (x\(top,lam (y\(tall (z\(top,tvar z)), tapp (var y) (tvar x)))))) T. 

? typeof [] (rcd []) (trcd []). 

? typeof [(tid_pair x top), (tid_pair s top)] (tlam (z\ (tvar x, lam (s'\ (tvar x ==> tvar s, lam (z'\ (tvar z, app (var s') (var z')))))))) T.

? ( patt (var y) (pvar (tvar t)) (pterm1 (x\ ( tvar t, var x))) ) ===> E.

? typeof [(id_pair y top), (tid_pair t top)] ( patt (var y) (pvar (tvar t)) (pterm1 (x\ ( tvar t, var x))) ) T.

(* ? typeof [] (lam (append\(top, var append))) T.	*)

(*? (app 
   (lam (b41\ (tall (x_\ (top,tall (s_\ (tvar x_,tall (z_\ (tvar x_,(tvar x_ ==> tvar s_) ==> tvar z_ ==> tvar s_)))))),
     tlam (b64\ (top,
       tlam (b87\ (tvar b64,
         tlam (b110\ (tvar b64,
           lam (b121\ (tvar b64 ==> tvar b87,
             lam (b132\ (tvar b110,
               app (app (tapp (tapp (tapp (
                 tlam (b979\ (top,
                   tlam (b1002\ (tvar b979,
                     tlam (b1025\ (tvar b979,
                       lam (b1036\ (tvar b979 ==> tvar b1002,
                         lam (b1047\ (tvar b1025,
                           app (var b1036) (app (app (tapp (tapp (tapp (
                             tlam (b1933\ (top,
                               tlam (b1956\ (tvar b1933,
                                 tlam (b1979\ (tvar b1933,
                                   lam (b1990\ (tvar b1933 ==> tvar b1956,
                                     lam (b2001\ (tvar b1979,
                                       app (var b1990) (app (app (tapp (tapp (tapp (
                                         tlam (b2999\ (top,
                                           tlam (b3022\ (tvar b2999,
                                             tlam (b3045\ (tvar b2999,
                                               lam (b3056\ (tvar b2999 ==> tvar b3022,
                                                 lam (b3067\ (tvar b3045,
                                                   app (app (tapp (tapp (tapp (
                                                     tlam (b3208\ (top,
                                                       tlam (b3231\ (tvar b3208,
                                                         tlam (b3254\ (tvar b3208,
                                                           lam (b3265\ (tvar b3208 ==> tvar b3231,
                                                             lam (b3276\ (tvar b3254,var b3276))))))))))) 
                                                     (tvar b2999)) (tvar b3022)) (tvar b3045)) (var b3056)) 
                                                       (app (app (tapp (tapp (tapp (
                                                         tlam (b5154\ (top,
                                                           tlam (b5177\ (tvar b5154,
                                                             tlam (b5200\ (tvar b5154,
                                                               lam (b5211\ (tvar b5154 ==> tvar b5177,
                                                                 lam (b5222\ (tvar b5200,var b5222))))))))))) (tvar b2999)) (tvar b3022)) (tvar b3045)) (var b3056)) (var b3067))))))))))))) (tvar b1933)) (tvar b1956)) (tvar b1979)) (var b1990)) (var b2001))))))))))))) (tvar b979)) (tvar b1002)) (tvar b1025)) (var b1036)) (var b1047))))))))))))) (tvar b64)) (tvar b87)) (tvar b87)) (var b121)) 
                   (app (app (tapp (tapp (tapp (var b41) (tvar b64)) (tvar b87)) (tvar b110)) (var b121)) (var b132))))))))))))))) 

   (app (app 
          (lam (b5342\ (tall (x_\ (top,tall (s_\ (tvar x_,tall (z_\ (tvar x_,(tvar x_ ==> tvar s_) ==> tvar z_ ==> tvar s_)))))),
            lam (b5353\ (tall (x_\ (top,tall (s_\ (tvar x_,tall (z_\ (tvar x_,(tvar x_ ==> tvar s_) ==> tvar z_ ==> tvar s_)))))),
              tlam (b5376\ (top,
                tlam (b5399\ (tvar b5376,
                  tlam (b5422\ (tvar b5376,
                    lam (b5433\ (tvar b5376 ==> tvar b5399,
                      lam (b5444\ (tvar b5422,
                        app (app (tapp (tapp (tapp (var b5342) (tvar b5376)) (tvar b5399)) (tvar b5399)) (var b5433)) 
                            (app (app (tapp (tapp (tapp (var b5353) (tvar b5376)) (tvar b5399)) (tvar b5422)) (var b5433)) (var b5444))))))))))))))))) 
          (tlam (b1905\ (top,
             tlam (b1928\ (tvar b1905,
               tlam (b1951\ (tvar b1905,
                 lam (b1962\ (tvar b1905 ==> tvar b1928,
                   lam (b1973\ (tvar b1951,
                     app (var b1962) (app (app (tapp (tapp (tapp (
                       tlam (b6676\ (top,
                         tlam (b6699\ (tvar b6676,
                           tlam (b6722\ (tvar b6676,
                             lam (b6733\ (tvar b6676 ==> tvar b6699,lam (b6744\ (tvar b6722,var b6744))))))))))) (tvar b1905)) (tvar b1928)) (tvar b1951)) (var b1962)) (var b1973)))))))))))))) 
        (app (var suc) (app (var suc) (app (var suc) (app (var suc) 
                                                          (tlam (b7167\ (top,
                                                             tlam (b7190\ (tvar b7167,
                                                               tlam (b7213\ (tvar b7167,
                                                                 lam (b7224\ (tvar b7167 ==> tvar b7190,
                                                                   lam (b7235\ (tvar b7213,var b7235))))))))))))))))) ---> E. *)

(* ? (app (lam (b41\(tall (x_\(top,tall (s_\(tvar x_,tall (z_\(tvar x_,(tvar x_ ==> tvar s_) ==> tvar z_ ==> tvar s_)))))),tlam (b64\(top,tlam (b87\(tvar b64,tlam (b110\(tvar b64,lam (b121\(tvar b64 ==> tvar b87,lam (b132\(tvar b110,app (app (tapp (tapp (tapp (tlam (b979\(top,tlam (b1002\(tvar b979,tlam (b1025\(tvar b979,lam (b1036\(tvar b979 ==> tvar b1002,lam (b1047\(tvar b1025,app (var b1036) (app (app (tapp (tapp (tapp (tlam (b1933\(top,tlam (b1956\(tvar b1933,tlam (b1979\(tvar b1933,lam (b1990\(tvar b1933 ==> tvar b1956,lam (b2001\(tvar b1979,app (var b1990) (app (app (tapp (tapp (tapp (tlam (b2999\(top,tlam (b3022\(tvar b2999,tlam (b3045\(tvar b2999,lam (b3056\(tvar b2999 ==> tvar b3022,lam (b3067\(tvar b3045,app (app (tapp (tapp (tapp (tlam (b3208\(top,tlam (b3231\(tvar b3208,tlam (b3254\(tvar b3208,lam (b3265\(tvar b3208 ==> tvar b3231,lam (b3276\(tvar b3254,var b3276))))))))))) (tvar b2999)) (tvar b3022)) (tvar b3045)) (var b3056)) (app (app (tapp (tapp (tapp (tlam (b5154\(top,tlam (b5177\(tvar b5154,tlam (b5200\(tvar b5154,lam (b5211\(tvar b5154 ==> tvar b5177,lam (b5222\(tvar b5200,var b5222))))))))))) (tvar b2999)) (tvar b3022)) (tvar b3045)) (var b3056)) (var b3067))))))))))))) (tvar b1933)) (tvar b1956)) (tvar b1979)) (var b1990)) (var b2001))))))))))))) (tvar b979)) (tvar b1002)) (tvar b1025)) (var b1036)) (var b1047))))))))))))) (tvar b64)) (tvar b87)) (tvar b87)) (var b121)) (app (app (tapp (tapp (tapp (var b41) (tvar b64)) (tvar b87)) (tvar b110)) (var b121)) (var b132))))))))))))))) (app (lam (b911\(tall (x_\(top,tall (s_\(tvar x_,tall (z_\(tvar x_,(tvar x_ ==> tvar s_) ==> tvar z_ ==> tvar s_)))))),tlam (b934\(top,tlam (b957\(tvar b934,tlam (b980\(tvar b934,lam (b991\(tvar b934 ==> tvar b957,lam (b1002\(tvar b980,app (app (tapp (tapp (tapp (tlam (b1905\(top,tlam (b1928\(tvar b1905,tlam (b1951\(tvar b1905,lam (b1962\(tvar b1905 ==> tvar b1928,lam (b1973\(tvar b1951,app (var b1962) (app (app (tapp (tapp (tapp (tlam (b6676\(top,tlam (b6699\(tvar b6676,tlam (b6722\(tvar b6676,lam (b6733\(tvar b6676 ==> tvar b6699,lam (b6744\(tvar b6722,var b6744))))))))))) (tvar b1905)) (tvar b1928)) (tvar b1951)) (var b1962)) (var b1973))))))))))))) (tvar b934)) (tvar b957)) (tvar b957)) (var b991)) (app (app (tapp (tapp (tapp (var b911) (tvar b934)) (tvar b957)) (tvar b980)) (var b991)) (var b1002))))))))))))))) (app (var suc) (app (var suc) (app (var suc) (app (var suc) (tlam (b7167\(top,tlam (b7190\(tvar b7167,tlam (b7213\(tvar b7167,lam (b7224\(tvar b7167 ==> tvar b7190,lam (b7235\(tvar b7213,var b7235))))))))))))))))) ---> E. *)
