%{

open Str
open String

let make_pattern p (b,e) =
   let reg = regexp "BLAH" in
   let rec repl l s = 
      match l with [] -> s
      |           (x::xs) -> (let s' = replace_first reg x s in repl xs s')
   in
      repl b p, e 

(* let remove_underscore s = (String.set s ((String.length s) - 1) ('a')) in *)
let remove_underscore s = s

%}

%token <string> LCID 
%token <string> UCID
%token <string> LABEL
%token VAR LAM APP TLAM TAPP RCD PROJ PATT
%token PVAR PRCD PTERM1 PTERM2 
%token TVAR TOP ARROW TALL TRCD 
%token LPAREN RPAREN COMMA LSQAURE RSQUARE SLASH EOF LSQUARE RSQUARE DOT

%right ARROW

%start main
%type <string> main 

%%

main: 
  term { "( " ^ $1 ^ " )" }
| typ { "( " ^ $1 ^ " )" }
;

term: 
  LPAREN VAR LCID RPAREN { $3 }
| VAR LCID { $2 }
| LAM LPAREN LCID SLASH LPAREN typ COMMA term RPAREN RPAREN { "lambda " ^ $3 ^ " : " ^ $6 ^ ". " ^ $8 }
| APP term term { $2 ^ " " ^ $3 }
| TLAM LPAREN LCID SLASH LPAREN typ COMMA term RPAREN RPAREN { "lambda " ^ (String.uppercase (remove_underscore $3)) ^ " <: " ^ $6 ^ ". " ^ $8 }
| TAPP term typ { $2 ^ " [" ^ $3 ^ "]" }
| RCD LSQUARE labeltermseq RSQUARE { "{ " ^ $3 ^ " }" }
| PROJ term DOT LABEL { $2 ^ "." ^ $4 }
| PATT LPAREN term RPAREN LPAREN pattern RPAREN LPAREN pterm RPAREN { let p,e2 = make_pattern $6 $9 in
                                                                     "let " ^ p ^ " = " ^ $3 ^ " in " ^ e2 }
| UCID { $1 }
| parenterm { $1 }
;

parenterm: 
  LPAREN term RPAREN { "(" ^ $2 ^ ")" }
;

typ: 
  TVAR LCID { String.uppercase (remove_underscore $2) }
| LPAREN TVAR LCID RPAREN { String.uppercase (remove_underscore $3) }
| TOP { "Top" }
| aterm { $1 }
| TALL LPAREN LCID SLASH LPAREN typ COMMA typ RPAREN RPAREN { "(All " ^ (String.uppercase (remove_underscore $3)) ^ " <: " ^ $6 ^ ". " ^ $8  ^ ")"}
| TRCD LSQUARE labeltypseq RSQUARE { "{ " ^ $3 ^ " }" }
| LPAREN typ RPAREN { "(" ^ $2 ^ ")" }
| UCID { $1 }
;

aterm: 
  typ ARROW typ { $1 ^ " -> " ^ $3 }
;

labeltermseq: 
  LPAREN LABEL COMMA term RPAREN { $2 ^ " = " ^ $4 }
| LPAREN LABEL COMMA term RPAREN COMMA labeltermseq { $2 ^ " = " ^ $4 ^ ", " ^ $7 }
;

labeltypseq: 
  LPAREN LABEL COMMA typ RPAREN { $2 ^ " = " ^ $4 }
| LPAREN LABEL COMMA typ RPAREN COMMA labeltypseq { $2 ^ " = " ^ $4 ^ ", " ^ $7 }
;

pattern: 
  PVAR typ { "BLAH : " ^ $2 }
| PRCD LSQUARE labelpattseq RSQUARE { "{ " ^ $3 ^ " }" }
;

labelpattseq: 
  LPAREN LABEL COMMA pattern RPAREN { $2 ^ " = " ^ $4 }
| LPAREN LABEL COMMA pattern RPAREN COMMA labelpattseq { $2 ^ " = " ^ $4 ^ ", " ^ $7 }
;

pterm: 
  PTERM1 LPAREN LCID SLASH LPAREN typ COMMA  term RPAREN RPAREN { [($3)], $8 }
| PTERM2 LPAREN LCID SLASH LPAREN typ COMMA pterm RPAREN RPAREN { let a,e = $8 in (($3)::a), e }
;

